import re
from pathlib import Path
from dataclasses import dataclass
import shutil
from typing import List, Callable, Tuple
from urllib.parse import unquote


@dataclass
class ImageText:
    """
    Data container for image.
    This class contains image file name, path and line number in markdown.
    """

    name: str
    line_no: int
    path: Path
    metadata: str
    s3_url: str | None = None

    def __post_init__(self):
        self.metadata = "" if not self.metadata else self.metadata


def _extract_image_from_no_path_patt(match_result: re.Match) -> dict:
    image_info = match_result.groupdict()
    image_name = image_info.get("name")
    assert image_name, "ImageText doesn't have path"
    return image_info


def _extract_image_from_path_patt(match_result: re.Match) -> dict | None:
    image_info = match_result.groupdict()
    image_path = image_info.get("path")
    assert image_path, "ImageText doesn't have path"
    if image_path[:4] in ("http", "https"):  # if link is external link
        return None

    try:
        image_info["name"] = unquote(Path(image_path).name)  # update imageText name
        return image_info

    except Exception:
        raise ValueError(f"{image_path} is invalid path")


def _extract_image_by_patt(
    func_patt_map: List[Tuple[Callable, str]],
    name_path_map: dict[str, Path],
    line: str,
    line_no: int,
) -> List[dict]:

    image_infos: List[dict] = []
    for func, patt in func_patt_map:
        if match_results := re.finditer(patt, line):  # match pattern
            for match_result in match_results:
                if image_info := func(match_result):  # run match func for patt
                    try:
                        image_info["path"] = name_path_map[image_info["name"]]
                        image_info["line_no"] = line_no
                        image_infos.append(image_info)
                    except KeyError:  # image file could be not exists
                        continue

    return image_infos


def extract_images_from_md(
    markdown_file_path: Path, name_path_map: dict[str, Path]
) -> List[ImageText]:
    """
    Extract images from md file

    Args:
        markdown_file_path (Path): markdown file path
        name_path_map (dict[str, Path]): image name path map {abc.png : /foo/abc.png}

    Returns:
        List[ImageText]: Image data in markdown
    """
    no_path_patt = r"!\[\[(?P<name>[^]]+\.(png|jpg|jpeg|gif))\|?(?P<metadata>[^]]+)?\]\]"  # group 1 = file_name, group 2 = foramt, group3 = imageText metadata
    path_patt = r"!\[(?P<metadata>[^]]+)?\]\((?P<path>[^)]+\.png|jpg|jpeg|gif\))"  # group 1 = metadata group 2 = file path

    func_patt_map: List[Tuple[Callable, str]] = [
        (_extract_image_from_no_path_patt, no_path_patt),
        (_extract_image_from_path_patt, path_patt),
    ]  # regex patt and match function

    images: List[ImageText] = []
    with markdown_file_path.open("r") as f:
        for line_no, line in enumerate(f):  # read lines
            image_infos = _extract_image_by_patt(
                func_patt_map, name_path_map, line, line_no
            )
            for image_info in image_infos:
                images.append(ImageText(**image_info))

    return images


def get_images_name_path_map(image_folder_path: Path) -> dict[str, Path]:
    """
    get all images in folder and create [name, Path] dict of all iamges

    Args:
        image_folder_path (Path): imageText folder path

    Returns:
        dict[str, Path]: {name, Path}
    """
    patt = r".*\.(png|jpg|jpeg|gif)$"
    name_path_map: dict[str, Path] = {}
    # search all subfolders
    for file_path in image_folder_path.rglob("**/*"):
        if re.search(patt, file_path.suffix):
            name_path_map[file_path.name] = file_path
    return name_path_map


def write_md_file(
    markdown_file_path: Path,
    output_folder_path: Path,
    uploaded_images: List[ImageText],
    is_overwrite: bool = False,
):
    """
    Write new .md that replace local file link to S3 url.
    Only replace imageText file link and other things are same
    Args:
        markdown_file_path (Path): md file path
        output_folder_path (Path): output file path
        link_replace_map (List[tuple[str, str]]): file link -> s3 url
    """
    out_file_path = output_folder_path.joinpath(markdown_file_path.name)
    with open(markdown_file_path, "r") as origin_file:  # open origin file
        with open(out_file_path, "w") as output_file:
            for line_no, line in enumerate(origin_file):
                for image in uploaded_images:
                    if line_no == image.line_no:
                        line = f"![{image.metadata}]({image.s3_url})\n"
                output_file.write(line)  # if no replace just copy line

    if is_overwrite:
        shutil.move(out_file_path, markdown_file_path)



import concurrent.futures
from pathlib import Path
from typing import Callable, List

from .markdown import (
    get_images_name_path_map,
    write_md_file,
    extract_images_from_md,
    ImageText,
)
from .s3 import S3


def _upload_images_from_md(
    s3: S3, markdown_path: Path, images: List[ImageText]
) -> List[ImageText]:
    """
    Generate local images path by using generator and put images to S3.
    Return sccessfully put image paths

    Args:
        s3 (S3): instance to control S3
        markdown_file_name (str): makrdown file name
        image_path_generator (Generator[Path, None, None]): yield image paths to upload

    Returns:
        List[Path]: sccessfully put image paths
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
        futures = []
        for image in images:
            if image.path:
                futures.append(
                    executor.submit(s3.put_image, markdown_path, image)
                )  # run upload by multithread

        uploaded_images: List[ImageText] = []  # put success image path list
        for future in concurrent.futures.as_completed(futures, timeout=60):
            uploaded_images.append(future.result())

    return uploaded_images


def create_obs3dian_runner(
    s3: S3,
    image_folder_path: Path,
    output_folder_path: Path,
    is_overwrite: bool = False,
) -> Callable:
    """
    Create runner fucntion object
    S3 controller and ouput_folder_path would not change before config

    Args:
        s3 (S3): S3 controller
        image_folder_path (Path): image folder path
        output_folder_path (Path): output folder path

    Returns:
        Callable: runner
    """

    name_path_map: dict[str, Path] = get_images_name_path_map(image_folder_path)

    def run(markdown_file_path: Path) -> None:
        """
        Run obs3dian command extract image paths and replace them by S3 URLs.

        Args:
            markdown_file_path (Path): mark down file path
        """
        images: List[ImageText] = extract_images_from_md(
            markdown_file_path, name_path_map
        )
        uploaded_images = _upload_images_from_md(s3, markdown_file_path, images)
        write_md_file(
            markdown_file_path, output_folder_path, uploaded_images, is_overwrite
        )  # write new md with S3 link
        return

    return run



import boto3
from botocore.exceptions import ClientError
import json
from pathlib import Path
from urllib import parse

from .markdown import ImageText


class S3:
    """
    Class to control S3
    this class can make bucket, set public access, put images in s3.
    To use this class you need AWS CLI profile which has permission to create public bucket and put images.
    """

    def __init__(
        self,
        bucket_name: str,
        profile_name: str | None = None,
        aws_access_key: str | None = None,
        aws_secret_key: str | None = None,
    ) -> None:

        if profile_name:
            self.session = boto3.Session(profile_name=profile_name)
        elif aws_access_key and aws_secret_key:
            self.session = boto3.Session(
                aws_access_key_id=aws_access_key,
                aws_secret_access_key=aws_secret_key,
                region_name="ap-northeast-2",
            )
        else:
            raise ValueError("Need profile name or Access Key & Secret Key")

        self.s3 = self.session.client("s3")
        self.bucket_name = bucket_name
        return

    def _check_bucket_exist(self) -> bool:
        try:
            self.s3.head_bucket(Bucket=self.bucket_name)  # check bucket is exists
            return True

        except ClientError as e:
            error_code = e.response["Error"]["Code"]
            if error_code == "404":  # if bucket not exists return false
                return False
            elif error_code == "403":  # 403 means bucket exists but no permission
                return True
            else:  # bad reqeust
                raise e

    def _put_public_access_policy(self) -> None:
        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "id-1",
                    "Action": ["s3:GetObject"],
                    "Effect": "Allow",
                    "Resource": f"arn:aws:s3:::{self.bucket_name}/*",
                    "Principal": "*",
                }
            ],
        }
        try:
            self.s3.put_bucket_policy(
                Bucket=self.bucket_name, Policy=json.dumps(policy)
            )  # put bucket public access

        except Exception as e:
            print("Can't allow public acess to bucket")
            raise e

    def create_bucket(self) -> bool:
        try:
            if self._check_bucket_exist():  # If bucket already exists
                return False

            bucket_name = self.bucket_name
            self.s3.create_bucket(
                CreateBucketConfiguration={"LocationConstraint": "ap-northeast-2"},
                Bucket=bucket_name,
                ObjectOwnership="ObjectWriter",
            )

            self.s3.put_public_access_block(
                Bucket=bucket_name,
                PublicAccessBlockConfiguration={
                    "BlockPublicAcls": False,
                    "IgnorePublicAcls": False,
                    "BlockPublicPolicy": False,
                    "RestrictPublicBuckets": False,
                },
            )
            self._put_public_access_policy()
            return True

        except ClientError as e:
            print("Error occured in create bucket")
            raise e

    def _get_image_url(self, markdown_path: Path, image_path: Path) -> str:
        region = self.session.region_name
        key = f"{markdown_path.stem} / {image_path.name}"
        s3_url = f"https://{self.bucket_name}.s3.{region}.amazonaws.com/{parse.quote(key)}"  # image uploaded url
        return s3_url

    def put_image(self, markdown_path: Path, image: ImageText) -> ImageText:
        try:
            self.s3.put_object(
                Bucket=self.bucket_name,
                Body=image.path.open("rb"),
                Key=f"{markdown_path.stem} / {image.name}",
                ContentType=f"image/{image.path.suffix}",
            )  # upload image
            s3_url = self._get_image_url(markdown_path, image.path)
            image.s3_url = s3_url
            return image

        except ClientError as e:
            print(f"Error Occured in uploading {image.path}")
            raise e
